/*
	\file		jailed.c
	\comment	Simple rootjailer.
	\author		johan@linkdata.se
*/

/*
	Copyright (c) 2004 Johan Lindh

	Permission is hereby granted, free of charge, to any person obtaining
	a copy of this software and associated documentation files (the
	"Software"), to deal in the Software without restriction, including
	without limitation the rights to use, copy, modify, merge, publish,
	distribute, sublicense, and/or sell copies of the Software, and to
	permit persons to whom the Software is furnished to do so, subject to
	the following conditions:

	The above copyright notice and this permission notice shall be
	included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
	MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
	BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
	ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
	CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
	SOFTWARE.
*/

#define VERSION "1.0.0"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>

#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <syslog.h>
#include <grp.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/fcntl.h>
#include <sys/stat.h>

#define HAVE_MKFIFO 1

typedef struct jail_child_
{
	pid_t			pid;			/**< The pid of the child process */
	uid_t			uid;			/**< The uid the child must run as */
	gid_t			gid;			/**< The gid the child must run as */
	time_t			started;		/**< When the child was started */
	int				sigchld_ign;	/**< Flag: ignore SIGCHLD */
	int				restart;		/**< Restart limit (in seconds) */
	int				no_pidfile;		/**< If nonzero, don't use a pidfile */
	char			pidfile[256];	/**< Where to write our (not the childs!) pid */
	char			root[256];		/**< Where to chroot() for the child */
	int				root_fd;		/**< File descriptor for above */
	char			user[96];		/**< Username of the child */
	char			program[96];	/**< Program name of the child (argv[0] without path or set with -n) */
	int				argc;			/**< argc for the child */
	char			**argv;			/**< argv for the child (max MAX_ARGC allowed) */
	int				envc;			/**< Number of environ strings in env */
	char			**env;			/**< Environment settings */
} jail_child;

jail_child child;
int debugging = 0;

const char *signame( int sig )
{

	switch( sig )
	{
#ifdef SIGHUP
	case SIGHUP: return "SIGHUP";
#endif
#ifdef SIGINT
	case SIGINT: return "SIGINT";
#endif
#ifdef SIGQUIT
	case SIGQUIT: return "SIGQUIT";
#endif
#ifdef SIGILL
	case SIGILL: return "SIGILL";
#endif
#ifdef SIGTRAP
	case SIGTRAP: return "SIGTRAP";
#endif
#ifdef SIGABRT
	case SIGABRT: return "SIGABRT";
#endif
#ifdef SIGEMT
	case SIGEMT: return "SIGEMT";
#endif
#ifdef SIGFPE
	case SIGFPE: return "SIGFPE";
#endif
#ifdef SIGKILL
	case SIGKILL: return "SIGKILL";
#endif
#ifdef SIGBUS
	case SIGBUS: return "SIGBUS";
#endif
#ifdef SIGSEGV
	case SIGSEGV: return "SIGSEGV";
#endif
#ifdef SIGSYS
	case SIGSYS: return "SIGSYS";
#endif
#ifdef SIGPIPE
	case SIGPIPE: return "SIGPIPE";
#endif
#ifdef SIGALRM
	case SIGALRM: return "SIGALRM";
#endif
#ifdef SIGTERM
	case SIGTERM: return "SIGTERM";
#endif
#ifdef SIGURG
	case SIGURG: return "SIGURG";
#endif
#ifdef SIGSTOP
	case SIGSTOP: return "SIGSTOP";
#endif
#ifdef SIGTSTP
	case SIGTSTP: return "SIGTSTP";
#endif
#ifdef SIGCONT
	case SIGCONT: return "SIGCONT";
#endif
#ifdef SIGCHLD
	case SIGCHLD: return "SIGCHLD";
#endif
#ifdef SIGTTIN
	case SIGTTIN: return "SIGTTIN";
#endif
#ifdef SIGTTOU
	case SIGTTOU: return "SIGTTOU";
#endif
#ifdef SIGIO
	case SIGIO: return "SIGIO";
#endif
#ifdef SIGXCPU
	case SIGXCPU: return "SIGXCPU";
#endif
#ifdef SIGXFSZ
	case SIGXFSZ: return "SIGXFSZ";
#endif
#ifdef SIGVTALRM
	case SIGVTALRM: return "SIGVTALRM";
#endif
#ifdef SIGPROF
	case SIGPROF: return "SIGPROF";
#endif
#ifdef SIGWINCH
	case SIGWINCH: return "SIGWINCH";
#endif
#ifdef SIGINFO
	case SIGINFO: return "SIGINFO";
#endif
#ifdef SIGUSR1
	case SIGUSR1: return "SIGUSR1";
#endif
#ifdef SIGUSR2
	case SIGUSR2: return "SIGUSR2";
#endif
	}
	return "<unknown>";
}

void cleanup( void )
{
	int i;
	if( child.argv )
	{
		for( i=0; i<child.argc; i++ )
		{
			if( child.argv[i] ) free( child.argv[i] );
			child.argv[i] = NULL;
		}
		free( child.argv );
		child.argc = 0;
		child.argv = NULL;
	}
	if( child.env )
	{
		for( i=0; i<child.envc; i++ )
		{
			if( child.env[i] ) free( child.env[i] );
			child.env[i] = NULL;
		}
		free( child.env );
		child.envc = 0;
		child.env = NULL;
	}
	if( child.root_fd != -1 )
	{
		close( child.root_fd );
		child.root_fd = -1;
	}
	if( !child.no_pidfile ) unlink( child.pidfile );

	return;
}

int redirect_io( void )
{
	char filename[256];
	int f;

	umask(0);

#if HAVE_MKFIFO
	/* redirect stdin */
	sprintf( filename, "/%s_stdin", child.program );
	mkfifo( filename, 0600 );
	f = open( filename, O_RDONLY | O_NONBLOCK, 0600 );
	if( f )
	{
		if( f != 0 ) dup2( f, 0 );
		if( f > 0 ) close( f );
	}
#endif

	/* redirect stdout */
	sprintf( filename, "/%s_stdout", child.program );
	f = open( filename, O_WRONLY | O_CREAT | O_APPEND, 0600 );
	if( f )
	{
		if( f != 1 ) dup2( f, 1 );
		if( f > 1 ) close( f );
	}

	/* redirect stderr */
	sprintf( filename, "/%s_stderr", child.program );
	f = open( filename, O_WRONLY | O_CREAT | O_APPEND, 0600 );
	if( f )
	{
		if( f != 2 ) dup2( f, 2 );
		if( f > 2 ) close( f );
	}

	return 0;
}

pid_t stop_child()
{
	int i;
	int status;
	pid_t pid = -1;

	if( child.pid != -1 )
	{
		child.sigchld_ign = 1;
		if( kill( child.pid, SIGTERM ) == 0 )
		{
			/* wait for it to die willingly */
			for( i=0; i<3*100; i++ )
			{
				usleep( 10000 );
				pid = waitpid( -1, &status, WNOHANG | WUNTRACED);
				if( pid == child.pid ) break;
			}
			if( pid == -1 )
			{
				syslog( LOG_WARNING, "(%s) child is ignoring SIGTERM, killing it", child.program );
				if( kill( child.pid, SIGKILL ) == 0 )
					pid = waitpid( -1, &status, WNOHANG | WUNTRACED);
				else
					syslog( LOG_WARNING, "(%s) kill(%d,SIGKILL) failed", child.program, child.pid );
			}
		}
		else
		{
			syslog( LOG_WARNING, "(%s) kill(%d,SIGTERM) failed", child.program, child.pid );
		}
		child.pid = -1;
		child.sigchld_ign = 0;
	}
	else
	{
		syslog( LOG_WARNING, "(%s) can't send signals to child PID -1", child.program );
	}

	if( pid == -1 )
	{
		syslog( LOG_WARNING, "(%s) failed to get child exit status", child.program );
	}

	return pid;
}

pid_t start_child()
{
	if( child.pid != -1 ) return child.pid;

	if( (child.pid = fork()) == 0 )
	{
		if( fchdir(child.root_fd)
			|| chroot(".")
			|| setgid(child.gid)
			|| setuid(child.uid)
			|| redirect_io()
			|| execve( child.argv[0], child.argv, child.env ) )
			syslog(LOG_ERR, "(%s) child failed to start in '%s': %m", child.program, child.root );

		exit(errno);
	}
	else if( child.pid == -1)
	{
		/* failed to fork off... */
		syslog(LOG_ERR, "(%s) can't fork(): %m", child.program );
	}

	/* set start time */
	time( &child.started );
	syslog(LOG_INFO, "(%s) child started, PID %d", child.program, child.pid );

	return child.pid;
}

void stop_and_exit( int s )
{
	syslog( LOG_WARNING, "(%s) received %s, stopping and exiting", child.program, signame(s) );
	stop_child();
	cleanup();
	exit(0);
	return;
}

void restart_child( int s )
{
	pid_t pid;
	int status;
	time_t now;

	signal( s, restart_child );
	pid = -1;

	if( s == SIGCHLD && child.sigchld_ign )
	{
		return;
	}

	/* on a HUP, stop our child rather than checking exit status */
	if( s == SIGHUP )
	{
		syslog( LOG_WARNING, "(%s) received %s, stopping and restarting", child.program, signame(s) );
		pid = stop_child();
	}
	else
	{
		pid = waitpid( -1, &status, WNOHANG | WUNTRACED);
		if( pid == child.pid )
		{
			if( WIFEXITED(status) )
			{
				syslog( LOG_WARNING, "(%s) child exited with code %d", child.program, WEXITSTATUS(status) );
			}
			else if( WIFSIGNALED(status) )
			{
				syslog( LOG_WARNING, "(%s) child terminated due to signal %s", child.program, signame(WTERMSIG(status)) );
			}
			else if( WIFSTOPPED(status) )
			{
				syslog( LOG_WARNING, "(%s) child stopped due to signal %s", child.program, signame(WSTOPSIG(status)) );
			}
			else
			{
				syslog( LOG_WARNING, "(%s) got a SIGCHLD, but child is running, so stopping it", child.program );
				pid = stop_child();
			}
		}
	}

	if( s == SIGCHLD && child.restart )
	{
		/* require at least child.restart seconds to have passed */
		/* since child.started */
		time( &now );
		if( now - child.started < child.restart )
		{
			/* died too soon, log and exit */
			syslog( LOG_ERR, "(%s) child exited within %d seconds, giving up", child.program, child.restart );
			cleanup();
			exit( 99 );
			return;
		}
	}

	/* restart the process */
	child.pid = -1;
	start_child();
	return;
}

/**
	Adds an argument to the arg list for the program.
*/
int add_arg( char *arg )
{
	char *arg_copy = NULL;
	char **new_argv;

	if( arg )
	{
		arg_copy = strdup( arg );
		if( arg_copy == NULL ) return 0;
	}

	new_argv = realloc( child.argv, sizeof(char*) * (child.argc+1) );
	if( new_argv == NULL ) return 0;
	child.argv = new_argv;
	child.argv[child.argc] = arg_copy;
	child.argc ++;

	return 1;
}

/**
	Converts a string to all uppercase until
	end-of-string or an equal sign is hit.
*/
void my_strupper( char *s )
{
	if( s )
	{
		while( *s && *s != '=' )
		{
			*s = toupper( *s );
			s ++;
		}
	}
}

/**
	Adds or replaces an environment string to the env list for the program.
*/
int add_env( char *arg )
{
	int i, len;
	char **new_env;
	char *arg_copy = NULL;
	char *p;
	char namebuf[256];

	/* see if it's there.. */
	if( arg )
	{
		arg_copy = strdup( arg );
		if( arg_copy == NULL ) return 0;
		my_strupper( arg_copy );

		p = strchr( arg_copy, '=' );
		if( p == NULL ) p = arg_copy;
		len = p - arg_copy;
		if( len > sizeof(namebuf)-1 ) len = sizeof(namebuf)-1;
		strncpy( namebuf, arg_copy, len );
		namebuf[len] = '\0';
		if( *namebuf == '\0' )
		{
			free( arg_copy );
			return 0;
		}
		for( i=0; i<child.envc; i++ )
		{
			if( !strncmp( namebuf, child.env[i], len ) )
			{
				free( child.env[i] );
				child.env[i] = arg_copy;
				return 1;
			}
		}
	}

	/* not there, add a new one */
	new_env = realloc( child.env, sizeof(char*) * (child.envc+1) );
	if( new_env == NULL )
	{
		if( arg_copy ) free( arg_copy );
		return 0;
	}
	child.env = new_env;
	child.env[child.envc] = arg_copy;
	child.envc ++;

	return 1;
}


/**
	Called from operating system.
*/
int main( int argc, char **argv )
{
	int i, n;
	char *p, *start;
	int pidfile = -1;
	char buf[1024];
	struct passwd *pwd = NULL;
	struct group *grp = NULL;

	memset( &child, 0, sizeof(child) );

	/* ensure we are running as root */
	if( geteuid() != 0 )
	{
		fprintf( stderr, "%s: must be run as root\n", argv[0] );
		return 1;
	}

	/* get the mandatory username/userid argument */
	/* else print basic usage instructions */
	if( argc > 1 )
	{
		strncpy( child.user, argv[1], sizeof(child.user) );
		child.user[sizeof(child.user)-1] = 0;
	}
	else
	{
		fprintf( stderr,
			"jailed version " VERSION ", Copyright (c) 2004 Johan Lindh\n"
			"This software comes with NO WARRANTY of any kind.\n"
			"\n"
			"usage: %s user [options] [program [args]]\n"
			"       user          A valid username or id to run as\n"
			"       -g<group>     Group name or id to run as\n"
			"       -r<secs>      Restart program if it fails after <secs> seconds\n"
			"       -c<chroot>    Where to chroot() to\n"
			"       -p<pidfile>   Where to write the pidfile (if blank, no pidfile)\n"
			"       -n<name>      Use this name rather than the programs in logs/files\n"
			"       -e<var>=<val> Set an environment variable\n"
			"       program       The program to start (relative new root)\n"
			"       args          The programs arguments\n"
			, argv[0] );
		return 2;
	}

	/* resolve username/id */
	if( isdigit(child.user[0]) )
	{
		child.uid = atoi( child.user );
		pwd = getpwuid( child.uid );
		if( pwd )
		{
			strncpy( child.user, pwd->pw_name?pwd->pw_name:"unknown", sizeof(child.user) );
			child.user[sizeof(child.user)-1] = 0;
		}
	}
	else
	{
		pwd = getpwnam( child.user );
		if( pwd ) child.uid = pwd->pw_uid;
	}

	/* bail out if no such user */
	if( pwd == NULL )
	{
		fprintf( stderr, "%s: no such user '%s'\n", argv[0], child.user );
		return 3;
	}

	/* fill in default values for the child */
	child.pid = -1;
	child.gid = pwd->pw_gid;
	child.no_pidfile = 0;
	child.pidfile[0] = '\0';
	strncpy( child.root, pwd->pw_dir?pwd->pw_dir:"/", sizeof(child.root) );
	child.root[sizeof(child.root)-1] = 0;
	child.restart = 5;
	child.program[0] = '\0';
	child.argc = 0;
	child.argv = NULL;
	child.envc = 0;
	child.env = NULL;
	add_env( "HOME=/" );
	add_env( "PATH=/bin" );
	sprintf( buf, "USER=%s", child.user );
	add_env( buf );

	/* parse arguments */
	for( i=2; i<argc; i++ )
	{
		if( child.argc == 0 )
		{
			/* program to start not yet seen; expect own arguments or program name */
			if( argv[i][0] == '-' )
			{
				switch( argv[i][1] )
				{
				case 'e':
					if( !add_env( argv[i]+2 ) )
					{
						fprintf( stderr, "%s: out of memory\n", argv[0] );
						return 255;
					}
					continue;
				case 'n':
					strncpy( child.program, argv[i]+2, sizeof(child.program) );
					child.program[ sizeof(child.program)-1 ] = 0;
					continue;
				case 'd':
					debugging = 1;
					continue;
				case 'g':
					/* specific gid */
					if( isdigit( argv[i][2] ) )
						child.gid = atoi( argv[i]+2 );
					else
					{
						grp = getgrnam( argv[i]+2 );
						if( grp == NULL )
						{
							fprintf( stderr, "%s: no such group '%s'\n", argv[0], argv[i]+2 );
							return 6;
						}
						child.gid = grp->gr_gid;
					}
					continue;
				case 'r':
					/* restart limit */
					if( isdigit( argv[i][2] ) )
						child.restart = atoi( argv[i]+2 );
					continue;
				case 'c':
					/* chroot */
					if( strlen( argv[i] ) > sizeof(child.root)-1 )
					{
						fprintf( stderr, "%s: sorry, I have a limit on %d characters for the chroot path\n",
							argv[0], sizeof(child.root)-1 );
						return 7;
					}
					strncpy( child.root, argv[i]+2, sizeof(child.root) );
					child.root[ sizeof(child.root)-1 ] = 0;
					continue;
				case 'p':
					/* pidfile */
					if( argv[i][2] )
					{
						child.no_pidfile = 0;
						strncpy( child.pidfile, argv[i]+2, sizeof(child.pidfile) );
						child.pidfile[ sizeof(child.pidfile)-1 ] = 0;
					}
					else
						child.no_pidfile = 1;
					continue;
				default:
					/* the program name started with a '-' */
					break;
				}
			}
			/* got program name */
			if( !add_arg( argv[i] ) )
			{
				fprintf( stderr, "%s: out of memory\n", argv[0] );
				return 255;
			}
		}
		else
		{
			/* program options */
			add_arg( argv[i] );
		}
	}

	/* if argc/argv for child not set, fill them in from user's shell */
	if( child.argc == 0 && pwd->pw_shell )
	{
		if( strlen( pwd->pw_shell ) > sizeof(buf)-1 )
		{
			fprintf( stderr, "%s: users shell line is too long! (>%d)\n", argv[0], sizeof(buf)-1 );
			return 9;
		}
		strcpy( buf, pwd->pw_shell );
		p = buf;

		while( *p )
		{
			while( *p && isspace(*p) ) p++;
			if( !*p ) break;

			start = p;
			while( *p && !isspace(*p) )
			{
				if( *p == '\"' )
				{
					p ++;
					while( *p && *p != '\"' ) p ++;
				}
				p ++;
			}
			if( isspace(*p) )
			{
				*p = 0;
				p ++;
			}
			if( !add_arg( start ) )
			{
				fprintf( stderr, "%s: out of memory\n", argv[0] );
				return 255;
			}
		}
	}

	/* add the terminating NULL arg */
	if( !add_arg( NULL ) )
	{
		fprintf( stderr, "%s: out of memory\n", argv[0] );
		return 255;
	}

	/* add the terminating NULL env */
	if( !add_env( NULL ) )
	{
		fprintf( stderr, "%s: out of memory\n", argv[0] );
		return 255;
	}

	/* set child.program to be the program name w/o any path */
	/* if not set with -n */
	if( child.program[0] == '\0' && child.argc > 0 && child.argv[0] )
	{
		p = strrchr( child.argv[0], '/' );
		if( p == NULL ) p = child.argv[0];
		else p ++;
		strncpy( child.program, p, sizeof(child.program) );
		child.program[ sizeof(child.program)-1 ] = 0;
	}

	/* ensure we have a reasonable setup */
	if( child.argc < 1 || child.user[0] == '\0' || child.program[0] == '\0' || child.root[0] == '\0' )
	{
		fprintf( stderr,
			"%s: this doesn't make sense to me:\n"
			"  user = '%s'\n"
			"  program = '%s'\n"
			"  chroot = '%s'\n"
			"  argc = %d\n"
			, argv[0], child.user, child.program, child.root, child.argc );
		return 10;
	}

	/* set pidfile, if not set using options */
	if( child.pidfile[0] == '\0' )
	{
		if( sprintf( child.pidfile, "/var/run/jailed_%s.pid", child.program) > sizeof(child.pidfile)-1 )
		{
			fprintf( stderr, "%s: internal: default pidfile name too long\n", argv[0] );
			return 255;
		}
	}

	/* open the root directory */
	child.root_fd = open( child.root, O_RDONLY );
	if( child.root_fd == -1 )
	{
		fprintf( stderr, "%s: can't access '%s'\n", argv[0], child.root );
		return 11;
	}

	/* open pidfile */
	if( !child.no_pidfile )
	{
		pidfile = open( child.pidfile, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL );
		if( pidfile == -1 )
		{
			fprintf( stderr, "%s: can't create pidfile %s, aborting\n", argv[0], child.pidfile );
			close( child.root_fd );
			return 11;
		}
	}

	if( debugging )
	{
		/* dump what we got */
		fprintf( stderr,
			"uid = %d\n"
			"gid = %d\n"
			"restart = %d\n"
			"pidfile = '%s'\n"
			"root = '%s'\n"
			"user = '%s'\n"
			"program = '%s'\n",
			child.uid,
			child.gid,
			child.restart,
			child.no_pidfile?"<not used>":child.pidfile,
			child.root,
			child.user,
			child.program
			);
		for( i=0; i<child.argc; i++ )
		{
			fprintf( stderr, "argv[%d] = '%s'\n",
				i, child.argv[i] );
		}
		for( i=0; i<child.envc; i++ )
		{
			fprintf( stderr, "env[%d] = '%s'\n",
				i, child.env[i] );
		}
	}

	/* set handlers */
	signal(SIGILL, stop_and_exit);
	signal(SIGSEGV, stop_and_exit);
	signal(SIGALRM, stop_and_exit);
	signal(SIGTERM, stop_and_exit);
	signal(SIGQUIT, stop_and_exit);
	signal(SIGCHLD, restart_child);
	signal(SIGHUP, restart_child);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGUSR1, SIG_IGN);
	signal(SIGUSR2, SIG_IGN);
	signal(SIGINT, SIG_IGN);

	daemon(0, 0);

	/* write my pid */
	if( pidfile != -1 )
	{
		n = sprintf( buf, "%d", getpid() );
		write( pidfile, buf, n );
		close( pidfile );
	}

	openlog("jailed", LOG_PID, LOG_DAEMON);

	if( start_child() == -1 )
	{
		return 12;
	}

	/* idle until exit()ed */
	for(;;)
	{
		usleep(100000);
	}

	return 0;
}

/* eof */
